import logger from '../logger.js'
import http2Client from 'http2-client'

const getProtocol = config => config.https ? 'https:' : 'http:'

const getHostFromServer = (server) => {
  const { address, family } = server.address()
  if (family === 'IPv6' && address === '::') {
    logger.info('Replace "::" with "0.0.0.0"')
    return '0.0.0.0'
  }
  if (family === 'IPv6') logger.warn(`Detected IPv6 address ${address}. This might not work.`)
  return address
}

const getPortFromServer = server => server.address()?.port || 80

const ping = (url, { timeout } = {}) => Promise.race([
  new Promise((resolve, reject) => {
    const { protocol, hostname, port } = url
    logger.info(`Send ping request to ${url}`)
    http2Client.request({
      protocol,
      hostname,
      port,
      path: '/ping',
      headers: {
        accept: '*/*',
        'ping-by-ox-proxy': true
      }
    }, res => {
      if (res.statusCode !== 200) return reject(new Error('Not found'))
      resolve()
    }).end()
  }),
  new Promise((resolve, reject) => setTimeout(() => reject(new Error('Request timed out')), timeout))
])

let devServerUrl

export const getDevServerURL = () => devServerUrl

export default function ({ frontends, httpServer, serverConfig }) {
  let isServerClosed = false
  httpServer.once('close', () => {
    isServerClosed = true
  })
  httpServer.once('listening', async () => {
    try {
      const protocol = getProtocol(serverConfig)
      const hostname = getHostFromServer(httpServer)
      const port = getPortFromServer(httpServer)
      const hostnames = [hostname, '127.0.0.1', 'localhost', '0.0.0.0']
        .filter((host, i, arr) => arr.indexOf(host) === i)

      for (const host of hostnames) {
        try {
          const url = new URL(`${protocol}//${host}:${port}`)
          logger.info(`Constructed url: "${url.href}". Let's see if its valid.`)
          await ping(url, { timeout: 50 })
          logger.info(`Detected dev-server on ${url.href}`)
          // TODO we might want to get additional configuration
          frontends.unshift({ target: url })
          devServerUrl = url
          return
        } catch (err) {
          if (isServerClosed) return
          logger.warn('Could not verify url')
        }
      }

      logger.fatal('Could not find a working local vite dev server.')
      if (!isServerClosed) process.exit(1)
    } catch (err) {
      logger.error(err)
    }
  })

  return async function (req, res, next) {
    // respond to ping requests
    if (req.headers['ping-by-ox-proxy']) {
      logger.info('Respond ping request')
      res.statusCode = 200
      res.end()
      return
    }

    next()
  }
}
